#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Pipes are the primary metaphor of the Meerschaum system.
You can interact with pipe data via `meerschaum.Pipe` objects.

If you are working with multiple pipes, it is highly recommended that you instead use
`meerschaum.utils.get_pipes` (available as `meerschaum.get_pipes`)
to create a dictionary of Pipe objects.

```
>>> from meerschaum import get_pipes
>>> pipes = get_pipes()
```

# Examples
For the below examples to work, `sql:remote_server` must be defined (check with `edit config`)
with correct credentials, as well as a network connection and valid permissions.

## Manually Adding Data
---

```
>>> from meerschaum import Pipe
>>> pipe = Pipe('csv', 'energy')
>>> 
>>> ### Columns only need to be defined if you're creating a new pipe.
>>> pipe.columns = { 'datetime' : 'time', 'id' : 'station_id' }
>>> 
>>> ### Create a Pandas DataFrame somehow,
>>> ### or you can use a dictionary of lists instead.
>>> df = pd.read_csv('data.csv')
>>> 
>>> pipe.sync(df)
```

## Registering a Remote Pipe
---

```
>>> from meerschaum import Pipe
>>> pipe = Pipe('sql:remote_server', 'energy')
>>> 
>>> pipe.attributes = {
...     'fetch' : {
...         'definition' : 'SELECT * FROM energy_table',
...     },
... }
>>> 
>>> ### Columns are a subset of attributes, so define columns
>>> ### after defining attributes.
>>> pipe.columns = { 'datetime' : 'time', 'id' : 'station_id' }
>>> 
>>> pipe.sync()
```

"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, Dict, Any, Union, InstanceConnector

class Pipe:
    """
    Access Meerschaum pipes via Pipe objects.

    Pipes are identified by the following:
        1. Connector keys (e.g. `sql:main`)
        2. Metric key (e.g. `weather`)
        3. Location (optional; e.g. `None`)

    A pipe's connector keys correspond to a data source, and when the pipe is synced,
    its `fetch` definition is evaluated and executed to produce new data.

    Alternatively, new data may be directly synced via `pipe.sync()`:

    ```
    >>> from meerschaum import Pipe
    >>> pipe = Pipe('csv', 'weather')
    >>> 
    >>> import pandas as pd
    >>> df = pd.read_csv('weather.csv')
    >>> pipe.sync(df)
    ```
    """

    from ._fetch import fetch
    from ._data import get_data, get_backtrack_data, get_rowcount
    from ._register import register
    from ._attributes import (
        attributes,
        parameters,
        columns,
        get_columns,
        get_columns_types,
        get_id,
        id,
    )
    from ._show import show
    from ._edit import edit, edit_definition
    from ._sync import sync, get_sync_time, exists
    from ._delete import delete
    from ._drop import drop
    from ._bootstrap import bootstrap

    def __init__(
        self,
        connector_keys : str,
        metric_key : str,
        location_key : Optional[str] = None,
        parameters : Optional[Dict[str, Any]] = None,
        columns : Optional[Dict[str, str]] = None,
        mrsm_instance : Optional[Union[str, InstanceConnector]] = None,
        instance : Optional[Union[str, InstanceConnector]] = None,
        debug : bool = False
    ):
        """
        :param connector_keys:
            Keys for the pipe's source connector.
            E.g. 'sql:main'

        :param metric_key:
            Label for the pipe's contents.
            E.g. 'weather'

        :param location_key:
            Label for the pipe's location.
            Defaults to None.

        :param parameters:
            Optionally set a pipe's parameters from the constructor,
            e.g. columns and other attributes.
            Defaults to None.

        :param columns:
            Subset of parameters for ease of use.
            Defaults to None.

        :param mrsm_instance:
            Connector keys for the Meerschaum instance where the pipe resides.
            Defaults to the preconfigured default instance.

        :param instance:
            Alias for `mrsm_instance`. If `mrsm_instance` is supplied, this value is ignored.
        """
        if location_key in ('[None]', 'None'):
            location_key = None
        self.connector_keys = connector_keys
        self.metric_key = metric_key
        self.location_key = location_key

        ### only set parameters if values are provided
        if parameters is not None:
            self._parameters = parameters

        if columns is not None:
            if self.__dict__.get('_parameters', None) is None:
                self._parameters = {}
            self._parameters['columns'] = columns

        ### NOTE: The parameters dictionary is {} by default.
        ###       A Pipe may be registered without parameters, then edited,
        ###       or a Pipe may be registered with parameters set in-memory first.
        from meerschaum.config import get_config
        _mrsm_instance = mrsm_instance if mrsm_instance is not None else instance
        if _mrsm_instance is None:
            _mrsm_instance = get_config('meerschaum', 'instance', patch=True)
        if not isinstance(_mrsm_instance, str):
            self._instance_connector = _mrsm_instance
            self.instance_keys = str(_mrsm_instance)
        else: ### NOTE: must be SQL or API Connector for this work
            self.instance_keys = _mrsm_instance

    @property
    def meta(self):
        """
        Simulate the MetaPipe model without importing FastAPI.
        """
        refresh = False
        if '_meta' not in self.__dict__:
            refresh = True
        elif self.parameters != self.__dict__['_meta']['parameters']:
            refresh = True

        if refresh:
            parameters = self.parameters
            if parameters is None:
                parameters = dict()
            self._meta = {
                'connector_keys' : self.connector_keys,
                'metric_key'     : self.metric_key,
                'location_key'   : self.location_key,
                'parameters'     : parameters,
            }
        return self._meta

    @property
    def instance_connector(self):
        """
        Return the connector for the instance on which the pipe resides.
        """
        if '_instance_connector' not in self.__dict__:
            from meerschaum.connectors.parse import parse_instance_keys
            conn = parse_instance_keys(self.instance_keys)
            if conn:
                self._instance_connector = conn
            else:
                return None
        return self._instance_connector

    @property
    def connector(self):
        """
        Return the pipe's data source connector.
        """
        if '_connector' not in self.__dict__:
            from meerschaum.connectors.parse import parse_instance_keys
            conn = parse_instance_keys(self.connector_keys)
            if conn:
                self._connector = conn
            else:
                return None
        return self._connector

    @property
    def sync_time(self):
        """
        Convenience function to get the pipe's latest datetime.
        """
        return self.get_sync_time()

    def __str__(self):
        """
        The Pipe's SQL table name. Converts the ':' in the connector_keys to an '_'.
        """
        name = f"{self.connector_keys.replace(':', '_')}_{self.metric_key}"
        if self.location_key is not None:
            name += f"_{self.location_key}"
        return name

    def __repr__(self):
        return str(self)

    def __getstate__(self):
        """
        Define the state dictionary (pickling).
        """
        state = {
            'connector_keys' : self.connector_keys,
            'metric_key' : self.metric_key,
            'location_key' : self.location_key,
            'parameters' : self.parameters,
            'mrsm_instance' :  self.instance_keys,
        }
        return state

    def __setstate__(self, _state : dict):
        """
        Read the state (unpickling).
        """
        self.__init__(**_state)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
"""
Functions for clearing elements
"""

from meerschaum.utils.typing import List, SuccessTuple, Any, Optional

def clear(
        action : Optional[List[str]] = None,
        **kw : Any
    ) -> SuccessTuple:
    """
    Clear pipes of their data, or clear the screen.
    """
    from meerschaum.utils.misc import choose_subaction
    if not action:
        from meerschaum.utils.formatting._shell import clear_screen
        return clear_screen(kw.get('debug', False)), ''

    options = {
        'pipes' : _clear_pipes,
    }
    return choose_subaction(action, options, **kw)

def _clear_pipes(
        action : Optional[List[str]] = None,
        **kw : Any
    ) -> SuccessTuple:
    """
    Clear pipes' data without dropping any tables.
    """
    return False, "Not implemented"

### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.utils.misc import choices_docstring as _choices_docstring
clear.__doc__ += _choices_docstring('clear')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Meerschaum wrapper around YAML libraries.

This is so switching between PyYAML and ruamel.yaml is smoother.
"""
_lib = None
_import_name = None
### The first import that is found will be used.
_yaml_imports = ['yaml', 'ruamel.yaml']
_yaml = None

class yaml:
    global _yaml, _import_name, _lib
    from meerschaum.utils.packages import attempt_import, all_packages, _import_module
    from meerschaum.utils.warnings import error
    for k in all_packages:
        if k in _yaml_imports:
            _import_name = k
            break
    if _import_name is None:
        error(f"No YAML libraries declared in meerschaum.packages.")

    try:
        _lib = _import_module(_import_name)
    except:
        _lib = attempt_import(_import_name, split=False, lazy=False, install=True)
    if _import_name == 'ruamel.yaml':
        _yaml = _lib.YAML()
    else:
        _yaml = _lib

    def safe_load(*args, **kw):
        from meerschaum.utils.misc import filter_keywords
        if _import_name == 'ruamel.yaml':
            return _yaml.load(*args, **filter_keywords(_yaml.load, **kw))
        return _yaml.safe_load(*args, **filter_keywords(_yaml.safe_load, **kw))

    def load(*args, **kw):
        from meerschaum.utils.misc import filter_keywords
        return _yaml.load(*args, **filter_keywords(_yaml.load, **kw))

    def dump(data, stream=None, **kw):
        from meerschaum.utils.misc import filter_keywords
        get_string = False
        if stream is None and _import_name == 'ruamel.yaml':
            stream = _lib.compat.StringIO()
            get_string = True
        result = _yaml.dump(data, stream, **filter_keywords(_yaml.dump, **kw))
        if get_string:
            return stream.getvalue()
        return result




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for interacting with the user.
"""

from __future__ import annotations
from meerschaum.utils.typing import Any, Union, Optional, Tuple, List

def prompt(
        question : str,
        icon : bool = True,
        default : Union[str, Tuple[str, str], None] = None,
        detect_password : bool = True,
        is_password : bool = False,
        wrap_lines : bool = True,
        noask : bool = False,
        **kw : Any
    ) -> str:
    """
    Ask the user a question and return the answer.
    Wrapper around `prompt_toolkit.prompt()` with modified behavior.
    For example, an empty string returns default instead of printing it for the user to delete
    (`prompt_toolkit` behavior).

    :param question:
        The question to print to the user.

    :param icon:
        If True, prepend the configured icon.

    :param default:
        If the response is '', return the default value.

    :param detect_password:
        If `True`, set the input method to a censored password box if the word `password`
        appears in the question.
        Defaults to `True`.

    :param is_password:
        If `True`, set the input method to a censored password box.
        May be overridden by `detect_password` unless `detect_password` is set to `False`.
        Defaults to `False`.

    :param wrap_lines:
        If `True`, wrap the text across multiple lines.
        Flag is passed onto `prompt_toolkit`.

    :param noask:
        If True, only print the question and return the default answer.
    """
    from meerschaum.utils.packages import attempt_import
    from meerschaum.utils.formatting import colored, ANSI, CHARSET
    from meerschaum.config import get_config
    from meerschaum.config.static import _static_config
    if not noask:
        prompt_toolkit = attempt_import('prompt_toolkit')
    question_config = get_config('formatting', 'question', patch=True)

    ### if a default is provided, append it to the question.
    default_answer = default
    if default is not None:
        question += f" (default: "
        if isinstance(default, tuple) and len(default) > 1:
            question += f"{default[0]} [{default[1]}]"
            default_answer = default[0]
        else:
            question += f"{default}"
        question += ")"
    #  if not question.endswith(': '):
        #  if question.endswith(':'): question += " "
        #  else: question += ": "

    ### detect password
    if (detect_password and 'password' in question.lower()) or is_password:
        kw['is_password'] = True
  
    ### Add the icon and only color the first line.
    lines = question.split('\n')
    first_line = lines[0]
    other_lines = '' if len(lines) <= 1 else '\n'.join(lines[1:])

    if ANSI:
        first_line = colored(first_line, **question_config['ansi']['rich'])

    _icon = question_config[CHARSET]['icon']
    question = (' ' + _icon + ' ') if icon and len(_icon) > 0 else ''
    question += first_line
    if len(other_lines) > 0:
        question += '\n' + other_lines
    question += ' '

    answer = (
        prompt_toolkit.prompt(
            prompt_toolkit.formatted_text.ANSI(question),
            wrap_lines = wrap_lines,
            **kw
        ) if not noask else ''
    )
    if noask:
        print(question)
    if answer == '' and default is not None:
        return default_answer
    return answer

def yes_no(
        question : str = '',
        options : Tuple[str, str] = ('y', 'n'),
        default : str = 'y',
        wrappers : Tuple[str, str] = ('[', ']'),
        icon : bool = True,
        yes : bool = False,
        noask : bool = False,
        interactive : bool = False,
        **kw : Any
    ) -> bool:
    """
    Print a question and prompt the user with a yes / no input.
    Returns True for 'yes', False for 'no'.

    :param question:
        The question to print to the user.

    :param options:
        The y/n options. The first is always considered `True`, and all options must be lower case.
        This behavior may be modifiable change in the future.

    :param default:
        The default option. Is represented with a capital to distinguish that it's the default.\
        E.g. [y/N] would return False by default.

    :param wrappers:
        Text to print around the '[y/n]' options.
        Defaults to ('[', ']').

    :param icon:
        If True, prepend the configured question icon.

    :param interactive:
        Not implemented. Was planning on using prompt_toolkit, but for some reason
        I can't figure out how to make the default selection 'No'.
    """
    from meerschaum.utils.warnings import error, warn
    from meerschaum.utils.formatting import ANSI, UNICODE
    from meerschaum.utils.packages import attempt_import

    default = options[0] if yes else default
    noask = yes or noask

    ending = f" {wrappers[0]}" + "/".join(
        [
            o.upper() if o.lower() == default.lower()
            else o.lower() for o in options
        ]
    ) + f"{wrappers[1]}"
    while True:
        try:
            answer = prompt(question + ending, icon=icon, detect_password=False, noask=noask)
            success = True
        except KeyboardInterrupt:
            success = False
        
        if not success:
            error(f"Error getting response. Aborting...", stack=False)
        if answer == "":
            answer = default

        if answer.lower() in options:
            break
        warn('Please enter a valid reponse.', stack=False)
    
    return answer.lower() == options[0].lower()

def choose(
        question : str,
        choices : List[str],
        default : Optional[str] = None,
        numeric : bool = True,
        icon : bool = True,
        warn : bool = True,
        noask : bool = False,
        **kw
    ) -> str:
    from meerschaum.utils.warnings import warn as _warn

    ### Handle empty choices.
    if len(choices) == 0:
        _warn(f"No available choices. Returning default value '{default}'.", stacklevel=3)
        return default

    ### Throw a warning if the default isn't a choice.
    if default is not None and default not in choices and warn:
        _warn(f"Default choice '{default}' is not contained in the choices {choices}. Setting numeric = False.", stacklevel=3)
        numeric = False

    _default = default
    _choices = choices
    if numeric:
        _choices = [str(i + 1) for i, c in enumerate(choices)]
        if default in choices:
            _default = str(choices.index(default) + 1)
        question += '\n'
        choices_digits = len(str(len(choices)))
        for i, c in enumerate(choices):
            question += f"  {i + 1}. " + (" " * (choices_digits - len(str(i + 1)))) + f"{c}\n"
        default_tuple = (_default, default) if default is not None else None
    else:
        default_tuple = default
        question += '\n'
        for c in choices:
            question += f"  - {c}\n"

    while True:
        answer = prompt(question, icon=icon, default=default_tuple, noask=noask, **kw)
        if answer in _choices or answer == default or noask:
            break
        _warn(f"Please pick a valid choice.", stack=False)

    if numeric:
        try:
            return choices[int(answer) - 1]
        except Exception as e:
            _warn(f"Could not cast answer '{answer}' to an integer.", stacklevel=3)

    return answer

def get_password(
        username : Optional[str] = None,
        minimum_length : Optional[int] = None,
        confirm : bool = True,
        **kw : Any
    ) -> str:
    """
    Prompt the user for a password.
    """
    from meerschaum.utils.warnings import warn
    while True:
        password = prompt(
            f"Password" + (f" for user '{username}':" if username is not None else ":"),
            is_password = True,
            **kw
        )
        if minimum_length is not None and len(password) < minimum_length:
            warn(
                "Password is too short. " +
                f"Please enter a password that is at least {minimum_length} characters.",
                stack = False
            )
            continue

        if not confirm:
            return password

        _password = prompt(
            f"Confirm password" + (f" for user '{username}':") if username is not None else ":",
            is_password = True,
            **kw
        )
        if password != _password:
            warn(f"Passwords do not match! Please try again.", stack=False)
            continue
        else:
            return password

def get_email(username : Optional[str] = None, allow_omit : bool = True, **kw : Any) -> str:
    """
    Prompt the user for an email and enforce that it's valid.

    :param username:
        Include an optional username to print.

    :param allow_omit:
        Allow the user to omit the email.
    """
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.misc import is_valid_email
    while True:
        email = prompt(
            f"Email for user" + (f" '{username}'" if username is not None else "") +
            (" (empty to omit): " if allow_omit else ": "),
            **kw
        )
        if (allow_omit and email == '') or is_valid_email(email):
            return email
        warn(f"Invalid email! Please try again.", stack=False)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Implement the Connector fetch() method
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, Union, Callable, Any

def fetch(
        self,
        pipe : meerschaum.Pipe.Pipe,
        begin : str = 'now',
        end : Optional[Union[datetime.datetime, str]] = None,
        chunk_hook : Optional[Callable[[pandas.DataFrame], Any]] = None,
        chunksize : Optional[int] = -1,
        debug : bool = False,
        **kw : Any
    ) -> Optional['pd.DataFrame']:
    """
    Execute the SQL definition and return a Pandas DataFrame.

    If pipe.columns['datetime'] and
        pipe.parameters['fetch']['backtrack_minutes'] are provided,
        append a `WHERE dt > begin` subquery.

    :param begin:
        Most recent datatime to search for data.
        If `backtrack_minutes` is provided, subtract `backtrack_minutes`.

    :param end:
        The latest datetime to search for data.
        If `end` is `None`, perform an unbounded search.

    :param pipe:
        Below are the various pipe parameters available to pipe.fetch.

        pipe.columns['datetime'] : str
            Name of the datetime column for the remote table.

        pipe.parameters['fetch'] : dict
            Parameters necessary to execute a query.
            See pipe.parameters['fetch'].

            pipe.parameters['fetch']['definition'] : str
                Raw SQL query to execute to generate the pandas DataFrame.

            pipe.parameters['backtrack_minutes'] : Union[int, float]
                How many minutes before `begin` to search for data.

    :param debug:
        Verbosity toggle.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn, error
    from meerschaum.connectors.sql._tools import sql_item_name, dateadd_str
    import datetime

    if 'columns' not in pipe.parameters or 'fetch' not in pipe.parameters:
        warn(f"Parameters for '{pipe}' must include 'columns' and 'fetch'", stack=False)
        return None

    dt_name = None
    if 'datetime' not in pipe.columns:
        warn(f"Missing datetime column for '{pipe}'. Will select all data instead.")
    else:
        dt_name = sql_item_name(pipe.get_columns('datetime'), self.flavor)

    instructions = pipe.parameters['fetch']

    try:
        definition = instructions['definition']
    except KeyError:
        error("Cannot fetch without a definition.", KeyError)

    if 'order by' in definition.lower() and 'over' not in definition.lower():
        error("Cannot fetch with an ORDER clause in the definition")

    da = None
    if dt_name:
        ### default: do not backtrack
        btm = 0
        if 'backtrack_minutes' in instructions:
            btm = instructions['backtrack_minutes']
        begin_da = dateadd_str(
            flavor=self.flavor, datepart='minute', number=(-1 * btm), begin=begin,
        )
        end_da = dateadd_str(
            flavor=self.flavor, datepart='minute', number=1, begin=end,
        ) if end else None

    meta_def = (
        _simple_fetch_query(pipe) if not pipe.columns.get('id', None)
        else _join_fetch_query(pipe, debug=debug, **kw)
    )

    if dt_name and (begin_da or end_da):
        meta_def += "\n" + ("AND" if 'where' in meta_def.lower() else "WHERE") + " "
        if begin_da:
            meta_def += f"definition.{dt_name} >= {begin_da}"
        if begin_da and end_da:
            meta_def += " AND "
        if end_da:
            meta_def += f"definition.{dt_name} <= {end_da}"

    df = self.read(meta_def, chunk_hook=chunk_hook, chunksize=chunksize, debug=debug)
    ### if sqlite, parse for datetimes
    if self.flavor == 'sqlite':
        from meerschaum.utils.misc import parse_df_datetimes
        df = parse_df_datetimes(df, debug=debug)
    return df

def _simple_fetch_query(pipe, debug: bool=False, **kw) -> str:
    """
    Build a fetch query from a pipe's definition.
    """
    definition = pipe.parameters['fetch']['definition']
    return f"WITH definition AS ({definition}) SELECT DISTINCT * FROM definition"

def _join_fetch_query(
        pipe,
        debug: bool=False,
        new_ids: bool=True,
        **kw
    ) -> str:
    """
    Build a fetch query based on the datetime and ID indices.
    """
    if not pipe.exists(debug=debug):
        return _simple_fetch_query(pipe, debug=debug, **kw)

    from meerschaum.connectors.sql._tools import sql_item_name, dateadd_str
    pipe_name = sql_item_name(str(pipe), pipe.connector.flavor)
    sync_times_table = str(pipe) + "_sync_times"
    sync_times_name = sql_item_name(sync_times_table, pipe.connector.flavor)
    id_name = sql_item_name(pipe.columns['id'], pipe.connector.flavor)
    dt_name = sql_item_name(pipe.columns['datetime'], pipe.connector.flavor)
    cols_types = pipe.get_columns_types(debug=debug)
    sync_times_query = f"""
    SELECT {id_name}, MAX({dt_name}) AS {dt_name}
    FROM {pipe_name}
    GROUP BY {id_name}
    """
    sync_times = pipe.connector.read(sync_times_query, debug=debug, silent=True)
    _sync_times_q = f",\n{sync_times_name} AS ("
    for _id, _st in sync_times.itertuples(index=False):
        _sync_times_q += (
            f"SELECT CAST('{_id}' AS " + cols_types[pipe.columns['id']] + f") AS {id_name}, "
            + dateadd_str(
                flavor=pipe.connector.flavor,
                begin=_st,
                datepart='minute',
                number=pipe.parameters.get('backtrack_minutes', 0)
            ) + " AS " + dt_name + "\nUNION ALL\n"
        )
    _sync_times_q = _sync_times_q[:(-1 * len('UNION ALL\n'))] + ")"

    definition = pipe.parameters['fetch']['definition']
    query = f"""
    WITH definition AS ({definition}){_sync_times_q}
    SELECT definition.*
    FROM definition
    LEFT OUTER JOIN {sync_times_name} AS st
      ON st.{id_name} = definition.{id_name}
    WHERE definition.{dt_name} > st.{dt_name}
    """ + (f"  OR st.{id_name} IS NULL" if new_ids else "")
    return query




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Fetch and manipulate Pipes' attributes
"""

from __future__ import annotations
from meerschaum.utils.typing import Tuple, Dict, SuccessTuple, Any

@property
def attributes(self) -> Optional[Dict[str, Any]]:
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn
    if '_attributes' not in self.__dict__:
        if self.id is None:
            return None
        self._attributes = self.instance_connector.get_pipe_attributes(self)
    return self._attributes

@property
def parameters(self) -> Optional[Dict[str, Any]]:
    if '_parameters' not in self.__dict__:
        if not self.attributes:
            return None
        self._parameters = self.attributes['parameters']
    return self._parameters

@parameters.setter
def parameters(self, parameters : Dict[str, Any]) -> None:
    self._parameters = parameters

@property
def columns(self):
    if not self.parameters:
        if '_columns' in self.__dict__:
            return self._columns
        return None
    if 'columns' not in self.parameters:
        return None
    return self.parameters['columns']

@columns.setter
def columns(self, columns : Dict[str, str]) -> None:
    if not self.parameters:
        self._columns = columns
    else:
        self._parameters['columns'] = columns

def get_columns(self, *args : str, error : bool = True) -> Tuple[str]:
    """
    Check if the requested columns are defined
    """
    from meerschaum.utils.warnings import error as _error, warn
    if not args:
        args = tuple(self.columns.keys())
    col_names = []
    for col in args:
        col_name = None
        try:
            col_name = self.columns[col]
            if col_name is None and error:
                _error(f"Please define the name of the '{col}' column for Pipe '{self}'.")
        except Exception as e:
            col_name = None
        if col_name is None and error:
            _error(f"Missing '{col}'" + f" column for Pipe '{self}'.")
        col_names.append(col_name)
    if len(col_names) == 1:
        return col_names[0]
    return tuple(col_names)

def get_columns_types(self, debug : bool = False) -> Optional[Dict[str, str]]:
    """
    Return a dictionary of a pipe's table's columns to their data types.

    E.g. An example dictionary for a small table.

    ```
    >>> {
    ...   'dt': 'TIMESTAMP WITHOUT TIMEZONE',
    ...   'id': 'BIGINT',
    ...   'val': 'DOUBLE PRECISION',
    ... }
    >>> 
    ```
    """
    return self.instance_connector.get_pipe_columns_types(self, debug=debug)

def get_id(self, **kw : Any) -> Optional[int]:
    """
    Fetch a pipe's ID from its instance connector.
    """
    return self.instance_connector.get_pipe_id(self, **kw)

@property
def id(self) -> Optional[int]:
    if not ('_id' in self.__dict__ and self._id):
        self._id = self.get_id()
    return self._id



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

import setuptools, sys
from setuptools import setup
from setuptools.command.install import install
cx_Freeze, Executable = None, None
#  if 'build_exe' in sys.argv or 'build_msi' in sys.argv or 'build_dmg' in sys.argv:
    #  try:
        #  import cx_Freeze
        #  from cx_Freeze import setup, Executable
    #  except ImportError:
        #  cx_Freeze, Executable = None, None

### read values from within the package
exec(open('meerschaum/config/_version.py').read())
exec(open('meerschaum/utils/packages/_packages.py').read())

class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        from meerschaum.config._edit import write_default_config
        from meerschaum.config._paths import CONFIG_DIR_PATH, PERMANENT_PATCH_DIR_PATH
        import os, shutil

        if CONFIG_DIR_PATH.exists():
            print(f"Found existing configuration in '{CONFIG_DIR_PATH}'")
            print(f"Moving to '{PERMANENT_PATCH_PATH}' and patching default configuration with existing configuration")
            shutil.move(CONFIG_DIR_PATH, PERMANENT_PATCH_DIR_PATH)
        else:
            print(f"Configuration not found: '{CONFIG_DIR_PATH}'")

extras = {}
### NOTE: package dependencies are defined in meerschaum.utils.packages._packages
for group in packages:
    extras[group] = [ install_name for import_name, install_name in packages[group].items() ]

with open('README.md', 'r', encoding='utf-8') as f:
    readme = f.read()

setup_kw_args = {
    'name'                          : 'meerschaum',
    'version'                       : __version__,
    'description'                   : 'Create and Manage Pipes with Meerschaum',
    'long_description'              : readme,
    'long_description_content_type' : 'text/markdown',
    'url'                           : 'https://meerschaum.io',
    'author'                        : 'Bennett Meares',
    'author_email'                  : 'bennett.meares@gmail.com',
    'maintainer_email'              : 'bennett.meares@gmail.com',
    'license'                       : 'Apache Software License 2.0',
    'packages'                      : setuptools.find_packages(),
    'install_requires'              : extras['required'],
    'extras_require'                : extras,
    'entry_points'                  : {
        'console_scripts'           : [
            'mrsm = meerschaum.__main__:main',
            'meerschaum = meerschaum.__main__:main',
        ],
    },
    #  'cmdclass'                      : {
        #  'install'                   : PostInstallCommand,
    #  },
    'zip_safe'                      : True,
    'package_data'                  : {'' : ['*.html', '*.css', '*.js', '*.png', '*.ico']},
    'python_requires'               : '>=3.7',
    'classifiers'                   : [
        "Development Status :: 4 - Beta",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Programming Language :: SQL",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Database",
    ],
}
#  if cx_Freeze is not None and Executable is not None:
    #  setup_kw_args['options'] = {
        #  'build_exe' : {
            #  'packages' : setuptools.find_packages(),
            #  'includes' : [
                #  'pip', 'venv', 'os', 'sys', 'importlib', 'pathlib',
            #  ],
            #  'optimize' : 0,
        #  },
    #  }
    #  setup_kw_args['executables'] = [
        #  Executable(
            #  "meerschaum/__main__.py",
            #  target_name = 'mrsm',
        #  )
    #  ]


setup(**setup_kw_args)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Directory of necessary packages

packages dictionary is structured in the following schema:
    {
        <group> : {
            <import_name> : <install_name>
        }
    }
"""

from __future__ import annotations
from meerschaum.utils.typing import Dict

packages : Dict[str, Dict[str, str]] = {
    'required' : {
        #  'setuptools' : 'setuptools',
        'wheel' : 'wheel>=0.34.2',
        'yaml' : 'PyYAML>=5.3.1',
        'cascadict' : 'cascadict>=0.8.4',
        'virtualenv' : 'virtualenv>=20.1.0',
    },
    'minimal' : {},
    'formatting' : {
        'pprintpp'          : 'pprintpp>=0.4.0',
        'asciitree'         : 'asciitree>=0.3.3',
        'typing_extensions' : 'typing_extensions>=3.7.4.3',
        'pygments'          : 'pygments>=2.7.2',
        'colorama'          : 'colorama>=0.4.3',
        'rich'              : 'rich>=9.8.0',
        'more_termcolor'    : 'more-termcolor>=1.1.3',
    },
    '_required' : {
        'pip'               : 'pip>=21.0.1',
        'update_checker'    : 'update-checker>=0.18.0',
        'semver'            : 'semver>=2.13.0',
        'dateutil'          : 'python-dateutil>=2.7.5',
        'requests'          : 'requests>=2.23.0',
        'binaryornot'       : 'binaryornot>=0.4.4',
        'dill'              : 'dill>=0.3.3',
        'pyvim'             : 'pyvim>=3.0.2',
        'aiofiles'          : 'aiofiles>=0.6.0',
        'packaging'         : 'packaging>=20.4',
        'prompt_toolkit'    : 'prompt-toolkit>=3.0.11',
        'more_itertools'    : 'more-itertools>=8.7.0',
        'daemoniker'        : 'daemoniker>=0.2.3',
        'watchgod'          : 'watchgod>=0.7',
        'pygtail'           : 'pygtail>=0.11.1',
        #  'nest_asyncio'      : 'nest-asyncio>=1.5.1',
    },
    'iot' : {
        'paho' : 'paho-mqtt>=1.5.1',
    },
    'drivers' : {
        'psycopg2'      : 'psycopg2-binary>=2.8.6',
        'pymysql'       : 'PyMySQL>=0.9',
        'aiomysql'      : 'aiomysql>=0.0.21',
        'cockroachdb'   : 'sqlalchemy-cockroachdb>=1.3.2',
        'duckdb'        : 'duckdb>=0.2.6',
        'duckdb_engine' : 'duckdb-engine>=0.1.3',
    },
    'cli' : {
        'pgcli'    : 'pgcli>=3.1.0',
        'mycli'    : 'mycli>=1.23.2',
        'litecli'  : 'litecli>=1.5.0',
        'mssqlcli' : 'mssqlcli>=1.0.4',
    },
    'stack' : {
        'docker'  : 'docker>=4.3.1',
        'compose' : 'docker-compose>=1.27.4',
    },
    'build' : {
        'cx_Freeze'   : 'cx_Freeze>=6.5.1',
        'PyInstaller' : 'pyinstaller>=5.0.dev0',
    },
    'dev-tools' : {
        'twine' : 'twine>=3.2.0',
        'tuna'     : 'tuna>=0.5.3',
        'snakeviz' : 'snakeviz>=2.1.0',
        'mypy' : 'mypy>=0.812',
        'pytest' : 'pytest>=6.2.2',
        'heartrate' : 'heartrate>=0.2.1',
        'pyheat' : 'py-heat>=0.0.6',
    },
    'setup' : {
    },
    'docs' : {
        'mkdocs'                      : 'mkdocs>=1.1.2',
        'pdoc'                        : 'pdoc3>=0.9.2',
        'mkdocs_material'             : 'mkdocs-material>=6.2.5',
        'mkdocs_autolinks_plugin'     : 'mkdocs-autolinks-plugin>=0.2.0',
        'mkdocs_awesome_pages_plugin' : 'mkdocs-awesome-pages-plugin>=2.5.0',
        'mkdocs_rss_plugin'           : 'mkdocs-rss-plugin>=0.16.1'
    },
    'portable' : {
        'pyreadline' : 'pyreadline3>=3.3; platform_system == "Windows"',
        'gnureadline' : 'gnureadline>=8.0.0; platform_system != "Windows"',
    },
    'extras' : {
        'cmd2' : 'cmd2>=1.4.0',
        'ruamel.yaml' : 'ruamel.yaml>=0.16.12',
        'pandasgui' : 'pandasgui>=0.2.9',      
        'modin' : 'modin[ray]>=0.8.3'
    },
}
packages['sql'] = {
    'numpy'      : 'numpy>=1.18.5',
    'pandas'     : 'pandas>=1.1.4',
    'joblib'     : 'joblib>=0.17.0',
    'sqlalchemy' : 'sqlalchemy>=1.4.17',
    'sqlalchemy_utils' : 'sqlalchemy-utils>=0.37.3',
    'databases'  : 'databases>=0.4.0',
    'aiosqlite'  : 'aiosqlite>=0.16.0',
    'asyncpg'    : 'asyncpg>=0.21.0',
}
packages['sql'].update(packages['drivers'])
packages['dash'] = {
    'dash'                      : 'dash>=1.20.0',
    'dash_bootstrap_components' : 'dash-bootstrap-components>=0.12.0',
    'dash_ace'                  : 'dash-ace>=0.2.1',
    'dash_extensions'           : 'dash-extensions>=0.0.51',
    'dash_daq'                  : 'dash-daq>=0.5.0',
    'ansi2html'                 : 'ansi2html>=1.6.0',
}
packages['api'] = {
    'uvicorn'       : 'uvicorn[standard]>=0.13.4',
    'websockets'    : 'websockets>=8.1',
    'fastapi'       : 'fastapi>=0.61.2',
    'fastapi_jwt_auth' : 'fastapi-jwt-auth>=0.5.0',
    #  'jinja2'        : 'jinja2>=2.11.2',
    'passlib'       : 'passlib>=1.7.4',
    'fastapi_login' : 'fastapi-login>=1.6.0',
    'multipart'     : 'python-multipart>=0.0.5',
    #  'capturer'                  : 'capturer>=3.0; platform_system != "Windows"',
}
packages['api'].update(packages['sql'])
packages['api'].update(packages['_required'])
packages['api'].update(packages['formatting'])
packages['api'].update(packages['dash'])

all_packages = dict()
for group, import_names in packages.items():
    all_packages.update(import_names)

skip_groups = {'docs', 'build', 'cli', 'dev-tools', 'portable', 'extras'}
full = list()
_full = dict()
for group, import_names in packages.items():
    ### omit 'cli' and 'docs' from 'full'
    if group in skip_groups:
        continue
    full += [ install_name for import_name, install_name in import_names.items() ]
    for import_name, install_name in import_names.items():
        _full[import_name] = install_name
packages['full'] = _full




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Default actions available to the mrsm CLI.
"""

from __future__ import annotations
from meerschaum.utils.typing import Callable, Any, Optional, Union, List
from meerschaum.utils.packages import get_modules_from_package
from meerschaum.utils.warnings import enable_depreciation_warnings
enable_depreciation_warnings(__name__)
_shell = None
_custom_actions = []

### build __all__ from other .py files in this package
import sys
__all__, modules = get_modules_from_package(
    sys.modules[__name__],
    names = True,
)

### build the actions dictionary by importing all
### functions that do not begin with '_' from all submodules
from inspect import getmembers, isfunction
actions = dict()
for module in modules:
    ### A couple important things happening here:
    ### 1. Find all functions in all modules in `actions` package
    ###     (skip functions that begin with '_')
    ### 2. Add them as members to the Shell class
    ###     - Original definition : meerschaum.actions.shell.Shell
    ###     - New definition      : meerschaum.actions.Shell
    ### 3. Populate the actions dictionary with function names and functions
    ###
    ### UPDATE:
    ### Shell modifications have been deferred to get_shell in order to improve lazy loading.

    actions.update(
        dict(
            [
                ### __name__ and new function pointer
                (ob[0], ob[1])
                    for ob in getmembers(module)
                        if isfunction(ob[1])
                            ### check that the function belongs to the module
                            and ob[0] == module.__name__.replace('_', '').split('.')[-1]
                            ### skip functions that start with '_'
                            and ob[0][0] != '_'
            ]
        )
    )

original_actions = actions.copy()
from meerschaum.actions._entry import _entry as entry

def get_shell(
        sysargs : List[str] = [],
        reload : bool = False,
        debug : bool = False
    ):
    """
    Lazy load the Shell
    """
    global _shell
    from meerschaum.utils.debug import dprint
    import meerschaum.actions.shell as shell_pkg

    if _shell is None or reload:
        if debug:
            dprint("Loading the shell...")

        if _shell is None:
            shell_pkg._insert_shell_actions()
            _shell = shell_pkg.Shell(actions, sysargs=sysargs)
        elif reload:
            _shell.__init__()
    return _shell

def get_subactions(action : str) -> Dict[str, Callable[[Any], Any]]:
    """
    Return a dictionary of an action's subactions.
    """
    import importlib, inspect
    subactions = {}
    try:
        action_module = importlib.import_module(f"meerschaum.actions.{action}")
    except ImportError:
        action_module = None
    if action_module is not None:
        for name, f in inspect.getmembers(action_module):
            if not inspect.isfunction(f):
                continue
            if action + '_' in name and not name.lstrip('_').startswith('complete'):
                _name = name.replace(action, '')
                _name = _name.lstrip('_')
                subactions[_name] = f
    return subactions

def get_completer(action : str) -> Optional[Callable[['Shell', str, str, int, int], List[str]]]:
    """
    Search for a custom completer function for an action.
    """
    import importlib, inspect
    try:
        action_module = importlib.import_module(f"meerschaum.actions.{action}")
    except ImportError:
        action_module = None
    if action_module is not None:
        for name, f in inspect.getmembers(action_module):
            if not inspect.isfunction(f):
                continue
            if 'complete_' + action in name:
                return f
    return None

def _get_parent_plugin(stacklevel : int = 1) -> Optional[str]:
    """
    Return the name of the calling plugin.
    If this function is called from outside a Meerschaum plugin, it will return None.
    """
    from meerschaum.config._paths import PLUGINS_RESOURCES_PATH
    import inspect, re
    parent_globals = inspect.stack()[stacklevel][0].f_globals
    parent_file = parent_globals.get('__file__', '')
    if str(PLUGINS_RESOURCES_PATH) not in parent_file:
        return None
    return parent_globals['__name__'].replace('plugins.', '').split('.')[0]

import meerschaum.plugins
plugins = meerschaum.plugins.import_plugins()
### Instruct pdoc to skip the `meerschaum.actions.plugins` subdirectory.
__pdoc__ = {'plugins' : False}
make_action = meerschaum.plugins.make_action
meerschaum.plugins.load_plugins()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Global pools that are joined on exit.
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional

pools = None

def get_pool(pool_class_name : str = 'ThreadPool', workers : Optional[int] = None):
    """
    If the requested pool does not exist, instantiate it here.
    Pools are joined and closed on exit.
    """
    global pools
    if pools is None:
        pools = dict()

    def build_pool(workers):
        global pools
        from meerschaum.utils.warnings import warn
        from meerschaum.utils.packages import attempt_import
        try:
            Pool = getattr(
                attempt_import('multiprocessing.pool', warn=False, venv=None, lazy=False),
                pool_class_name
            )
        except Exception as e:
            warn(e, stacklevel=3)
            Pool = getattr(
                attempt_import('multiprocessing.pool', warn=False, venv=None, lazy=False),
                'ThreadPool'
            )

        if workers is None:
            from multiprocessing import cpu_count
            workers = cpu_count()
        try:
            pools[pool_class_name] = Pool(workers)
        except Exception as e:
            pools[pool_class_name] = None

    if pool_class_name not in pools or pools.get(pool_class_name, None) is None:
        build_pool(workers)

    if (
        pools[pool_class_name] is not None
        and pools[pool_class_name]._state not in ('RUN', 0)
    ):
        try:
            pools[pool_class_name].close()
        except Exception as e:
            pass
        del pools[pool_class_name]
        build_pool(workers)

    return pools[pool_class_name]

def get_pools():
    """
    Return the global pools dictionary.
    """
    global pools
    if pools is None:
        pools = dict()
    return pools



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Functions for getting Pipe data
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, Dict, Any

def get_data(
        self,
        begin : Optional[datetime.datetime] = None,
        end : Optional[datetime.datetime] = None,
        params : Optional[Dict[str, Any]] = None,
        refresh : bool = False,
        debug : bool = False,
        **kw : Any
    ) -> Optional[pandas.DataFrame]:
    """
    Get a pipe's data from the instance connector.

    :param begin:
        Lower bound datetime to begin searching for data (inclusive).
        Translates to a `WHERE` clause like `WHERE datetime >= begin`.
        Defaults to `None`.

    :param end:
        Upper bound datetime to stop searching for data (inclusive).
        Translates to a `WHERE` clause like `WHERE datetime <= end`.
        Defaults to `None`.

    :param params:
        Filter the retrieved data by a dictionary of parameters.
        E.g. to retrieve data for only certain values of `id`,
        the `params` dictionary would look like the following:
        
        ```
        >>> params = {
        ...   'id' : [1, 2, 3],
        ... }
        >>> 
        ```

    :param refresh:
        If True, skip local cache and directly query the instance connector.
        Currently has no effect (until caching features are merged into the stable release).
        Defaults to `True`.

    :param debug:
        Verbosity toggle.
        Defaults to `False`.
    """
    if refresh or True: ### TODO remove `or True`
        self._data = self.instance_connector.get_pipe_data(
            pipe = self,
            begin = begin,
            end = end,
            params = params,
            debug = debug,
            **kw
        )
    ### TODO caching / sync logic
    return self._data

def get_backtrack_data(
        self,
        backtrack_minutes : int = 0,
        begin : 'datetime.datetime' = None,
        debug : bool = False,
        **kw : Any
    ) -> Optional['pd.DataFrame']:
    """
    Get the most recent data from the instance connector as a Pandas DataFrame.

    :param backtrack_minutes:
        How many minutes from `begin` to select from.
        Defaults to 0. This may return a few rows due to a rounding quirk.

    :param begin:
        The starting point from which to search for data.
        If begin is None (default), use the most recent observed datetime
        (AKA sync_time).

        E.g. begin = 02:00

        ```

        Search this region.           Ignore this, even if there's data.
        /  /  /  /  /  /  /  /  /  |
        -----|----------|----------|----------|----------|----------|
           00:00      01:00      02:00      03:00      04:00      05:00

        ```

    :param debug: Verbosity toggle.
    """
    return self.instance_connector.get_backtrack_data(
        pipe = self,
        begin = begin,
        backtrack_minutes = backtrack_minutes,
        debug = debug,
        **kw
    )

def get_rowcount(
        self,
        begin : Optional['datetime.datetime'] = None,
        end : Optional['datetime.datetime'] = None,
        remote : bool = False,
        params : Optional[Dict[str, Any]] = None,
        debug : bool = False
    ) -> Optional[int]:
    """
    Get a Pipe's cached or remote rowcount.

    :param begin:
        Count rows where datetime > begin.

    :param end:
        Count rows where datetime <= end.

    :param remote:
        Count rows from a pipe's remote source.
        NOTE: This is experimental!

    :param debug: Verbosity toggle.
    """
    connector = self.instance_connector if not remote else self.connector
    try:
        return connector.get_pipe_rowcount(
            self, begin=begin, end=end, remote=remote, params=params, debug=debug
        )
    except AttributeError:
        if remote:
            return None
    from meerschaum.utils.warnings import warn
    warn(f"Failed to get a rowcount for pipe '{self}'.")
    return None



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Flavor-specific SQL tools.
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, Dict, Any, Union

test_queries = {
    'default' : 'SELECT 1',
    'oracle' : 'SELECT 1 FROM DUAL',
    'informix' : 'SELECT COUNT(*) FROM systables',
    'hsqldb' : 'SELECT 1 FROM INFORMATION_SCHEMA.SYSTEM_USERS',
}

def test_connection(
        self,
        **kw : Any
    ) -> Union[bool, None]:
    """
    Block until a connection to the SQL database is made.
    """
    import warnings
    from meerschaum.utils.misc import retry_connect
    _default_kw = {'max_retries' : 1, 'retry_wait' : 0, 'warn' : False,}
    _default_kw.update(kw)
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', 'Could not')
        return retry_connect(**_default_kw)

def get_distinct_col_count(
        col : str,
        query : str,
        connector : Optional[meerschaum.connectors.sql.SQLConnector] = None,
        debug : bool = False
    ) -> Optional[int]:
    """
    Returns the number of distinct items in a column of a SQL query.

    :param col:
        The column in the query to count.

    :param query:
        The SQL query to count from.

    :param connector:
        The SQLConnector to execute the query.
    """
    
    if connector is None:
        from meerschaum import get_connector
        connector = get_connector('sql')

    _col_name = sql_item_name(col, connector.flavor)

    _meta_query = f"""
    WITH src AS ( {query} ),
    dist AS ( SELECT DISTINCT {_col_name} FROM src )
    SELECT COUNT(*) FROM dist"""

    result = connector.value(_meta_query, debug=debug)
    try:
        return int(result)
    except Exception as e:
        return None

def sql_item_name(s : str, flavor : str) -> str:
    """
    Parse SQL items depending on the flavor
    """
    if flavor in {'timescaledb', 'postgresql', 'cockroachdb'}:
        s = pg_capital(str(s))
    elif flavor == 'sqlite':
        s = "\"" + str(s) + "\""
    return str(s)

def pg_capital(s : str) -> str:
    """
    If string contains a capital letter, wrap it in double quotes

    returns: string
    """
    if '"' in s:
        return s
    needs_quotes = False
    for c in str(s):
        if ord(c) < ord('a') or ord(c) > ord('z'):
            if not c.isdigit():
                needs_quotes = True
                break
    if needs_quotes:
        return '"' + s + '"'
    return s

def build_where(
        parameters : Dict[str, Any],
        connector : Optional[meerschaum.connectors.sql.SQLConnector] = None
    ) -> str:
    """
    Build the WHERE clause based on the input criteria
    """
    if connector is None:
        from meerschaum import get_connector
        connector = get_connector('sql')
    where = ""
    leading_and = "\n    AND "
    for key, value in parameters.items():
        _key = sql_item_name(key, connector.flavor)
        ### search across a list (i.e. IN syntax)
        if isinstance(value, list):
            where += f"{leading_and}{_key} IN ("
            for item in value:
                where += f"'{item}', "
            where = where[:-2] + ")"
            continue

        ### search a dictionary
        elif isinstance(value, dict):
            import json
            where += (f"{leading_and}CAST({_key} AS TEXT) = '" + json.dumps(value) + "'")
            continue

        where += f"{leading_and}{_key} " + ("IS NULL" if value is None else f"= '{value}'")
    if len(where) > 1:
        where = "\nWHERE\n    " + where[len(leading_and):]
    return where

def dateadd_str(
        flavor : str = 'postgresql',
        datepart : str = 'day',
        number : Union[int, float] = -1,
        begin : Union[str, datetime.datetime] = 'now'
    ) -> str:
    """
    Generate a DATEADD clause depending on database flavor.
    This function is pretty fragile / complex, so I may depreciate
    it in favor of a pure-Python or ORM solution.

    :param flavor:
        SQL database flavor, e.g. postgresql, sqlite.
        Currently supported flavors:
        - postgresql
        - timescaledb
        - cockroachdb
        - duckdb
        - mssql
        - mysql
        - mariadb
        - sqlite
        - oracle

    :param datepart:
        Which part of the date to modify. Supported values* (*AFAIK).
        - year
        - month
        - day
        - hour
        - minute
        - second

    :param number:
        How many units to add to the date part.

    :param begin:
        Base datetime to which to add dateparts.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.packages import attempt_import
    import datetime
    dateutil = attempt_import('dateutil')
    if not begin:
        return None
    begin_time = None
    if not isinstance(begin, datetime.datetime):
        try:
            begin_time = dateutil.parser.parse(begin)
        except Exception:
            begin_time = None
    else:
        begin_time = begin

    da = ""
    if flavor in ('postgresql', 'timescaledb', 'cockroachdb'):
        if begin == 'now':
            begin = "CAST(NOW() AT TIME ZONE 'utc' AS TIMESTAMP)"
        elif begin_time:
            begin = f"CAST('{begin}' AS TIMESTAMP)"
        da = begin + f" + INTERVAL '{number} {datepart}'"
    elif flavor == 'duckdb':
        if begin == 'now':
            begin = 'NOW()'
        elif begin_time:
            begin = f"CAST('{begin}' AS TIMESTAMP)"
        da = begin + f" + INTERVAL '{number} {datepart}'"
    elif flavor in ('mssql',):
        if begin == 'now':
            begin = "GETUTCDATE()"
        elif begin_time:
            begin = f"CAST('{begin}' AS DATETIME)"
        da = f"DATEADD({datepart}, {number}, {begin})"
    elif flavor in ('mysql', 'mariadb'):
        if begin == 'now':
            begin = "UTC_TIMESTAMP()"
        elif begin_time:
            begin = f'"{begin}"'
        da = f"DATE_ADD({begin}, INTERVAL {number} {datepart})"
    elif flavor == 'sqlite':
        da = f"datetime('{begin}', '{number} {datepart}')"
    elif flavor == 'oracle':
        if begin == 'now':
            begin = str(
                datetime.datetime.utcnow().strftime('%Y:%m:%d %M:%S.%f')
            )
        elif begin_time:
            begin = str(begin.strftime('%Y:%m:%d %M:%S.%f'))
        dt_format = 'YYYY-MM-DD HH24:MI:SS.FF'
        da = f"TO_TIMESTAMP('{begin}', '{dt_format}') + INTERVAL '{number}' {datepart}"
    return da

def table_exists(
        table: str,
        connector: Optional[meerschaum.connectors.sql.SQLConnector] = None,
        debug: bool = False,
    ) -> bool:
    """
    Check if a table exists.
    """
    if connector is None:
        from meerschaum import get_connector
        connector = get_connector('sql')

    table_name = sql_item_name(table, connector.flavor)
    ### default: select no rows. NOTE: this might not work for Oracle
    q = f"SELECT COUNT(*) FROM {table_name} WHERE 1 = 0"
    if connector.flavor in ('timescaledb', 'postgresql'):
        q = f"SELECT to_regclass('{table_name}')"
    elif connector.flavor == 'mssql':
        q = f"SELECT OBJECT_ID('{table_name}')"
    elif connector.flavor in ('mysql', 'mariadb'):
        q = f"SHOW TABLES LIKE '{table_name}'"
    elif connector.flavor == 'sqlite':
        q = f"SELECT name FROM sqlite_master WHERE name='{table}'"
    exists = connector.value(q, debug=debug, silent=True) is not None
    return exists

def get_sqlalchemy_table(
        table: str,
        connector: Optional[meerschaum.connectors.sql.SQLConnector] = None,
        debug: bool = False,
    ) -> sqlalchemy.Table:
    """
    Return a sqlalchemy Table bound to a SQLConnector's engine.
    """
    if connector is None:
        from meerschaum import get_connector
        connector = get_connector('sql')

    from meerschaum.connectors.sql.tables import get_tables
    from meerschaum.utils.packages import attempt_import
    tables = get_tables(mrsm_instance=connector, debug=debug)
    sqlalchemy = attempt_import('sqlalchemy')
    if str(table) not in tables:
        tables[str(table)] = sqlalchemy.Table(
            str(table),
            connector.metadata,
            autoload_with = connector.engine
        )
    return tables[str(table)]




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Implement WebSockets for the Meerschaum API via FastAPI.
"""

import time, datetime, uuid
from meerschaum.api import (
    app, get_api_connector, get_uvicorn_config, debug, fastapi, endpoints
)
from meerschaum.config.static import _static_config
from meerschaum.utils.typing import Optional

#  _websocket_endpoint = _static_config()['api']['endpoints']['websocket']
_websocket_endpoint = endpoints['websocket']

websockets = {}
sessions = {}

@app.websocket(_websocket_endpoint)
async def websocket_endpoint(
        websocket : fastapi.WebSocket,
    ):
    """
    Communicate with the Web Interface over a websocket.
    """
    global websockets
    await websocket.accept()
    initial_data = await websocket.receive_json()
    session_id = initial_data['session-id']
    now = datetime.datetime.utcnow()
    join_msg = ""
    await websocket.send_text(join_msg)
    websockets[session_id] = websocket
    ### NOTE: remove the below statement!
    websockets['debug'] = websocket
    while True:
        try:
            data = await websocket.receive_text()
            await websocket.send_text(str(datetime.datetime.utcnow()))
        except fastapi.WebSocketDisconnect:
            del websockets[session_id]
            break



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Formatting functions for the interactive shell
"""

def make_header(
        message : str,
        ruler : str = '─',
    ) -> str:
    """
    Format a message string with a ruler.
    Length of the ruler is the length of the longest word

    Example:
        'My\nheader' -> 'My\nheader\n──────'
    """

    from meerschaum.utils.formatting import ANSI, UNICODE, colored
    if not UNICODE:
        ruler = '-'
    words = message.split('\n')
    max_length = 0
    for w in words:
        length = len(w)
        if length > max_length:
            max_length = length

    s = message + "\n"
    for i in range(max_length):
        s += ruler
    return s

def clear_screen(debug : bool = False) -> bool:
    """
    Clear the terminal window of all text. If ANSI is enabled,
    print the ANSI code for clearing. Otherwise, execute `clear` or `cls`.
    """
    from meerschaum.utils.formatting import ANSI, get_console
    from meerschaum.utils.debug import dprint
    print("", end="", flush=True)
    if debug:
        dprint("Skipping screen clear.")
        return True
    if ANSI:
        if get_console() is not None:
            get_console().clear()
            print("", end="", flush=True)
            return True
        clear_string, reset_string = '\033c', '\033[0m'
        print(clear_string + reset_string, end="")
        print("", end="", flush=True)
        return True
    ### ANSI support is disabled, try system level instead
    import platform, subprocess
    command = 'clear' if platform.system() != "Windows" else "cls"
    rc = subprocess.call(command, shell=False)
    return rc == 0



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
"""
This module contains SQLConnector functions for executing SQL queries.
"""

from __future__ import annotations
from meerschaum.utils.typing import (
    Union, Mapping, List, Dict, SuccessTuple, Optional, Any, Sequence, Iterable, Callable
)

from meerschaum.utils.debug import dprint
from meerschaum.utils.warnings import warn

### database flavors that can use bulk insert
_bulk_flavors = {'postgresql', 'timescaledb'}
### flavors that do not support chunks
_disallow_chunks_flavors = {'duckdb'}

def read(
        self,
        query_or_table : Union[str, sqlalchemy.Query],
        params : Optional[Dict[str, Any], List[str]] = None,
        chunksize : Optional[int] = -1,
        chunk_hook : Optional[Callable[[pandas.DataFrame], Any]] = None,
        as_hook_results : bool = False,
        chunks : Optional[int] = None,
        as_chunks : bool = False,
        as_iterator: bool = False,
        silent : bool = False,
        debug : bool = False,
        **kw : Any
    ) -> Optional[
            Union[
                pandas.DataFrame,
                List[pandas.DataFrame],
                List[Any],
            ]
        ]:
    """
    Read a SQL query or table into a pandas dataframe.

    :param query_or_table:
        The SQL query (sqlalchemy Query or string) or name of the table from which to select.

    :param params:
        `List` or `Dict` of parameters to pass to `pandas.read_sql()`.
        See the pandas documentaion for more information:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.read_sql.html

    :param chunksize:
        How many chunks to read at a time. `None` will read everything in one large chunk.
        Defaults to system configuration.

    :param chunk_hook:
        Hook function to execute once per chunk.
        NOTE: `as_iterator` MUST be False (default).
        E.g. Write and reading chunks intermittently. See `--sync-chunks` for an example.

    :param as_hook_results:
        If `True`, return a `List` of the outputs of the hook function.
        Only applicable if `chunk_hook` is not None.
        NOTE: `as_iterator` MUST be False (default).
        Defaults to False.

    :param chunks:
        How many chunks to retrieve and return into a single dataframe.

    :param as_chunks:
        If `True`, return a list of DataFrames. Otherwise return a single DataFrame.
        Defaults to False.

    :param as_iterator:
        If `True`, return the pandas DataFrame iterator.
        `chunksize` must not be None (falls back to 1000 if so),
            and hooks are not called in this case.
        Defaults to False.

    :param silent:
        If `True`, don't raise warnings in case of errors.
        Defaults to `False`.

    """
    if chunks is not None and chunks <= 0:
        return []
    from meerschaum.connectors.sql._tools import sql_item_name
    from meerschaum.utils.packages import attempt_import, import_pandas
    pd = import_pandas()
    sqlalchemy = attempt_import("sqlalchemy")
    chunksize = chunksize if chunksize != -1 else self.sys_config.get('chunksize', None)
    if chunksize is None and as_iterator:
        if not silent and self.flavor not in _disallow_chunks_flavors:
            warn(f"An iterator may only be generated if chunksize is not None. Falling back to a chunksize of 1000.", stacklevel=3)
        chunksize = 1000

    ### NOTE: A bug in duckdb_engine does not allow for chunks.
    if chunksize is not None and self.flavor in _disallow_chunks_flavors:
        chunksize = None

    if debug:
        import time
        start = time.time()
        dprint(query_or_table)
        dprint(f"Fetching with chunksize: {chunksize}")

    ### This might be sqlalchemy object or the string of a table name.
    ### We check for spaces and quotes to see if it might be a weird table.
    if (
        ' ' not in str(query_or_table)
        or (
            ' ' in str(query_or_table)
            and str(query_or_table).startswith('"')
            and str(query_or_table).endswith('"')
        )
    ):
        query_or_table = sql_item_name(str(query_or_table), self.flavor)
        if debug:
            dprint(f"Reading from table {query_or_table}")
        formatted_query = str(sqlalchemy.text("SELECT * FROM " + str(query_or_table)))
    else:
        try:
            formatted_query = str(sqlalchemy.text(query_or_table))
        except:
            formatted_query = query_or_table

    try:
        chunk_generator = pd.read_sql(
            formatted_query,
            self.engine,
            params = params,
            chunksize = chunksize
        )
    except Exception as e:
        import inspect
        if debug:
            dprint(f"Failed to execute query:\n\n{query_or_table}\n\n")
        if not silent:
            warn(str(e), stacklevel=3)

        return None

    chunk_list = []
    read_chunks = 0
    chunk_hook_results = []
    if chunksize is None:
        chunk_list.append(chunk_generator)
    elif as_iterator:
        return chunk_generator
    else:
        for chunk in chunk_generator:
            if chunk_hook is not None:
                chunk_hook_results.append(
                    chunk_hook(chunk, chunksize=chunksize, debug=debug, **kw)
                )
            chunk_list.append(chunk)
            read_chunks += 1
            if chunks is not None and read_chunks >= chunks:
                break

    ### If no chunks returned, read without chunks
    ### to get columns
    if len(chunk_list) == 0:
        chunk_list.append(
            pd.read_sql(
                formatted_query,
                self.engine
            )
        )

    ### call the hook on any missed chunks.
    if chunk_hook is not None and len(chunk_list) > len(chunk_hook_results):
        for c in chunk_list[len(chunk_hook_results):]:
            chunk_hook_results.append(
                chunk_hook(c, chunksize=chunksize, debug=debug, **kw)
            )

    ### chunksize is not None so must iterate
    if debug:
        end = time.time()
        dprint(f"Fetched {len(chunk_list)} chunks in {round(end - start, 2)} seconds.")

    if as_hook_results:
        return chunk_hook_results
    
    ### Skip `pd.concat()` if `as_chunks` is specified.
    if as_chunks:
        for c in chunk_list:
            c.reset_index(drop=True, inplace=True)
        return chunk_list

    return pd.concat(chunk_list).reset_index(drop=True)

def _read_duckdb(query : str, engine : sqlalchemy.Engine, ):
    """
    Implement the `pandas.read_sql()` method for duckdb.
    """
    pass

def value(
        self,
        query : str,
        *args : Any,
        use_pandas : bool = False,
        **kw : Any
    ) -> Any:
    """
    Return a single value from a SQL query
    (index a DataFrame at [0, 0])
    """
    if use_pandas:
        try:
            return self.read(query, *args, **kw).iloc[0, 0]
        except Exception as e:
            return None
    try:
        result = self.exec(query, *args, **kw)
        return result.first()[0]
    except Exception as e:
        return None

def execute(
        self,
        *args : Any,
        **kw : Any
    ) -> Optional[sqlalchemy.engine.result.resultProxy]:
    return self.exec(*args, **kw)

def exec(
        self,
        query : str,
        *args : Any,
        silent : bool = False,
        debug : bool = False,
        **kw : Any
    ) -> Optional[sqlalchemy.engine.result.resultProxy]:
    """
    Execute SQL code and return success status. e.g. calling stored procedures.

    Wrapper for self.engine.connect() and connection.execute().

    If inserting data, please use bind variables to avoid SQL injection!
    """
    from meerschaum.utils.packages import attempt_import
    sqlalchemy = attempt_import("sqlalchemy")
    if debug:
        dprint("Executing query:\n" + f"{query}")

    connection = self.engine.connect()
    transaction = connection.begin()
    try:
        result = connection.execute(query, *args, **kw)
        transaction.commit()
    except Exception as e:
        if debug:
            dprint(f"Failed to execute query:\n\n{query}\n\n{e}")
        if not silent:
            warn(str(e))
        result = None
        transaction.rollback()
    finally:
        connection.close()

    return result

def to_sql(
        self,
        df : pandas.DataFrame,
        name : str = None,
        index : bool = False,
        if_exists : str = 'replace',
        method : str = "",
        chunksize : int = -1,
        silent : bool = False,
        debug : bool = False,
        as_tuple : bool = False,
        **kw
    ) -> Union[bool, SuccessTuple]:
    """
    Upload a DataFrame's contents to the SQL server

    :param df:
        The DataFrame to be uploaded

    :param name:
        The name of the table to be created

    :param index:
        If True, creates the DataFrame's indices as columns (default False)

    :param if_exists:
        ['replace', 'append', 'fail']
        Drop and create the table ('replace') or append if it exists
        ('append') or raise Exception ('fail').
        (default 'replace')

    :param method:
        None or multi. Details on pandas.to_sql

    :param as_tuple:
        If True, return a (success_bool, message) tuple instead of a bool

    **kw : keyword arguments
        Additional arguments will be passed to the DataFrame's `to_sql` function
    """
    import time
    from meerschaum.utils.warnings import error
    if name is None:
        error("Name must not be None to submit to the SQL server")

    from meerschaum.connectors.sql._tools import sql_item_name

    ### resort to defaults if None
    if method == "":
        if self.flavor in _bulk_flavors:
            method = psql_insert_copy
        else:
            ### Should default to 'multi'.
            method = self.sys_config.get('create_engine', {}).get('method', None)
    chunksize = chunksize if chunksize != -1 else self.sys_config.get('chunksize', None)

    success, msg = False, "Default to_sql message"
    start = time.time()
    if debug:
        msg = f"Inserting {len(df)} rows with chunksize: {chunksize}..."
        print(msg, end="", flush=True)

    ### filter out non-pandas args
    import inspect
    to_sql_params = inspect.signature(df.to_sql).parameters
    to_sql_kw = dict()
    for k, v in kw.items():
        if k in to_sql_params:
            to_sql_kw[k] = v

    try:
        df.to_sql(
            name = name,
            con = self.engine,
            index = index,
            if_exists = if_exists,
            method = method,
            chunksize = chunksize,
            **to_sql_kw
        )
        success = True
    except Exception as e:
        if not silent:
            warn(str(e))
        success, msg = None, str(e)

    end = time.time()
    if success:
        msg = f"It took {round(end - start, 2)} seconds to sync {len(df)} rows to {name}."

    if debug:
        print(f" done.", flush=True)
        dprint(msg)

    if as_tuple:
        return success, msg
    return success

def psql_insert_copy(
        table : pandas.io.sql.SQLTable,
        conn : Union[sqlalchemy.engine.Engine, sqlalchemy.engine.Connection],
        keys : Sequence[str],
        data_iter : Iterable[Any]
    ) -> None:
    """
    Execute SQL statement inserting data

    :param table : pandas.io.sql.SQLTable
    conn : sqlalchemy.engine.Engine or sqlalchemy.engine.Connection
    keys : list of str
        Column names
    data_iter : Iterable that iterates the values to be inserted
    """
    import csv
    from io import StringIO

    from meerschaum.connectors.sql._tools import sql_item_name

    # gets a DBAPI connection that can provide a cursor
    dbapi_conn = conn.connection
    with dbapi_conn.cursor() as cur:
        s_buf = StringIO()
        writer = csv.writer(s_buf)
        writer.writerows(data_iter)
        s_buf.seek(0)

        columns = ', '.join('"{}"'.format(k) for k in keys)
        if table.schema:
            table_name = '{}.{}'.format(
                sql_item_name(table.schema, 'postgresql'),
                sql_item_name(table.name, 'postgresql')
            )
        else:
            table_name = sql_item_name(table.name, 'postgresql')

        sql = 'COPY {} ({}) FROM STDIN WITH CSV'.format(
            table_name, columns
        )
        cur.copy_expert(sql=sql, file=s_buf)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Return version information
"""

from meerschaum.api import app, endpoints

@app.get(endpoints['version'])
def get_api_version():
    """
    Get the Meerschaum API version
    """
    from meerschaum.api import __version__ as version
    return version

@app.get(endpoints['version'] + "/mrsm")
def get_meerschaum_version():
    """
    Get the Meerschaum instance version
    """
    from meerschaum import __version__ as version
    return version



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Implement the prompt_toolkit Completer base class.
"""

from __future__ import annotations
from prompt_toolkit.completion import Completer, Completion
from meerschaum.utils.typing import Optional
from meerschaum.actions import get_shell, get_completer
from meerschaum.actions.arguments import parse_line

from meerschaum.utils.packages import attempt_import
prompt_toolkit = attempt_import('prompt_toolkit', lazy=False, install=True)

class ShellCompleter(Completer):
    """
    Implement the `prompt_toolkit` Completer to use the built-in `complete_` methods
    and the `cmd` completer system.
    """
    def get_completions(self, document, completer_event):
        """
        Bridge the built-in cmd completer with the `prompt_toolkit` completer system.
        """
        shell_actions = [a[3:] for a in dir(get_shell()) if a.startswith('do_')]
        yielded = []
        for i, a in enumerate(shell_actions):
            poss = get_shell().complete(document.text, i)
            if not poss:
                break
            yield Completion(poss, start_position=(-1 * len(poss)))
            yielded.append(poss)

        if yielded:
            return

        action = document.text.split(' ')[0]
        possibilities = []
        if action and f'complete_{action}' in dir(get_shell()):
            possibilities = getattr(
                get_shell(), f'complete_{action}'
            )(document.text.split(' ')[-1], document.text, 0, 0)

        if possibilities:
            for suggestion in possibilities:
                yield Completion(
                    suggestion, start_position=(-1 * len(document.text.split(' ')[-1]))
                )
            return

        yield Completion('')

